/* DO NOT EDIT - THIS FILE IS AUTOMATICALLY GENERATED */
/*
** License Applicability. Except to the extent portions of this file are
** made subject to an alternative license as permitted in the SGI Free
** Software License B, Version 1.1 (the "License"), the contents of this
** file are subject only to the provisions of the License. You may not use
** this file except in compliance with the License. You may obtain a copy
** of the License at Silicon Graphics, Inc., attn: Legal Services, 1600
** Amphitheatre Parkway, Mountain View, CA 94043-1351, or at:
** 
** http://oss.sgi.com/projects/FreeB
** 
** Note that, as provided in the License, the Software is distributed on an
** "AS IS" basis, with ALL EXPRESS AND IMPLIED WARRANTIES AND CONDITIONS
** DISCLAIMED, INCLUDING, WITHOUT LIMITATION, ANY IMPLIED WARRANTIES AND
** CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, FITNESS FOR A
** PARTICULAR PURPOSE, AND NON-INFRINGEMENT.
** 
** Original Code. The Original Code is: OpenGL Sample Implementation,
** Version 1.2.1, released January 26, 2000, developed by Silicon Graphics,
** Inc. The Original Code is Copyright (c) 1991-2000 Silicon Graphics, Inc.
** Copyright in any portions created by third parties is as indicated
** elsewhere herein. All Rights Reserved.
** 
** Additional Notice Provisions: This software was created using the
** OpenGL(R) version 1.2.1 Sample Implementation published by SGI, but has
** not been independently verified as being compliant with the OpenGL(R)
** version 1.2.1 Specification.
*/

#ifdef HAVE_DIX_CONFIG_H
#include <dix-config.h>
#endif

#include "glxserver.h"
#include "glxext.h"
#include "g_disptab.h"
__GLXdispatchSingleProcPtr __glXSingleTable[__GLX_SINGLE_TABLE_SIZE] = {
	__glXNoSuchSingleOpcode,
	__glXRender,
	__glXRenderLarge,
	__glXCreateContext,
	__glXDestroyContext,
	__glXMakeCurrent,
	__glXIsDirect,
	__glXQueryVersion,
	__glXWaitGL,
	__glXWaitX,
	__glXCopyContext,
	__glXSwapBuffers,
	__glXUseXFont,
	__glXCreateGLXPixmap,
	__glXGetVisualConfigs,
	__glXDestroyGLXPixmap,
	__glXVendorPrivate,
	__glXVendorPrivateWithReply,
	__glXQueryExtensionsString,
	__glXQueryServerString,
	__glXClientInfo,
	__glXGetFBConfigs,
	__glXCreatePixmap,
	__glXDestroyGLXPixmap, /* glXDestroyPixmap */
	__glXCreateNewContext,
	__glXNoSuchSingleOpcode, /* glXQueryContext */
	__glXMakeContextCurrent,
	__glXNoSuchSingleOpcode, /* glXCreatePbuffer */
	__glXNoSuchSingleOpcode, /* glXDestroyPbuffer */
	__glXNoSuchSingleOpcode, /* glXGetDrawableAttributes */
	__glXNoSuchSingleOpcode, /* glXChangeDrawableAttributes */
	__glXNoSuchSingleOpcode, /* glXCreateWindow */
	__glXNoSuchSingleOpcode, /* glXDestroyWindow */
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXDisp_NewList,
	__glXDisp_EndList,
	__glXDisp_DeleteLists,
	__glXDisp_GenLists,
	__glXDisp_FeedbackBuffer,
	__glXDisp_SelectBuffer,
	__glXDisp_RenderMode,
	__glXDisp_Finish,
	__glXDisp_PixelStoref,
	__glXDisp_PixelStorei,
	__glXDisp_ReadPixels,
	__glXDisp_GetBooleanv,
	__glXDisp_GetClipPlane,
	__glXDisp_GetDoublev,
	__glXDisp_GetError,
	__glXDisp_GetFloatv,
	__glXDisp_GetIntegerv,
	__glXDisp_GetLightfv,
	__glXDisp_GetLightiv,
	__glXDisp_GetMapdv,
	__glXDisp_GetMapfv,
	__glXDisp_GetMapiv,
	__glXDisp_GetMaterialfv,
	__glXDisp_GetMaterialiv,
	__glXDisp_GetPixelMapfv,
	__glXDisp_GetPixelMapuiv,
	__glXDisp_GetPixelMapusv,
	__glXDisp_GetPolygonStipple,
	__glXDisp_GetString,
	__glXDisp_GetTexEnvfv,
	__glXDisp_GetTexEnviv,
	__glXDisp_GetTexGendv,
	__glXDisp_GetTexGenfv,
	__glXDisp_GetTexGeniv,
	__glXDisp_GetTexImage,
	__glXDisp_GetTexParameterfv,
	__glXDisp_GetTexParameteriv,
	__glXDisp_GetTexLevelParameterfv,
	__glXDisp_GetTexLevelParameteriv,
	__glXDisp_IsEnabled,
	__glXDisp_IsList,
	__glXDisp_Flush,
	__glXDisp_AreTexturesResident,
	__glXDisp_DeleteTextures,
	__glXDisp_GenTextures,
	__glXDisp_IsTexture,
	__glXDisp_GetColorTable,
	__glXDisp_GetColorTableParameterfv,
	__glXDisp_GetColorTableParameteriv,
	__glXDisp_GetConvolutionFilter,
	__glXDisp_GetConvolutionParameterfv,
	__glXDisp_GetConvolutionParameteriv,
	__glXDisp_GetSeparableFilter,
	__glXDisp_GetHistogram,
	__glXDisp_GetHistogramParameterfv,
	__glXDisp_GetHistogramParameteriv,
	__glXDisp_GetMinmax,
	__glXDisp_GetMinmaxParameterfv,
	__glXDisp_GetMinmaxParameteriv,
};

__GLXdispatchRenderProcPtr __glXRenderTable[] = {
	__glXNoSuchRenderOpcode,
	__glXDisp_CallList,
	__glXDisp_CallLists,
	__glXDisp_ListBase,
	__glXDisp_Begin,
	__glXDisp_Bitmap,
	__glXDisp_Color3bv,
	__glXDisp_Color3dv,
	__glXDisp_Color3fv,
	__glXDisp_Color3iv,
	__glXDisp_Color3sv,
	__glXDisp_Color3ubv,
	__glXDisp_Color3uiv,
	__glXDisp_Color3usv,
	__glXDisp_Color4bv,
	__glXDisp_Color4dv,
	__glXDisp_Color4fv,
	__glXDisp_Color4iv,
	__glXDisp_Color4sv,
	__glXDisp_Color4ubv,
	__glXDisp_Color4uiv,
	__glXDisp_Color4usv,
	__glXDisp_EdgeFlagv,
	__glXDisp_End,
	__glXDisp_Indexdv,
	__glXDisp_Indexfv,
	__glXDisp_Indexiv,
	__glXDisp_Indexsv,
	__glXDisp_Normal3bv,
	__glXDisp_Normal3dv,
	__glXDisp_Normal3fv,
	__glXDisp_Normal3iv,
	__glXDisp_Normal3sv,
	__glXDisp_RasterPos2dv,
	__glXDisp_RasterPos2fv,
	__glXDisp_RasterPos2iv,
	__glXDisp_RasterPos2sv,
	__glXDisp_RasterPos3dv,
	__glXDisp_RasterPos3fv,
	__glXDisp_RasterPos3iv,
	__glXDisp_RasterPos3sv,
	__glXDisp_RasterPos4dv,
	__glXDisp_RasterPos4fv,
	__glXDisp_RasterPos4iv,
	__glXDisp_RasterPos4sv,
	__glXDisp_Rectdv,
	__glXDisp_Rectfv,
	__glXDisp_Rectiv,
	__glXDisp_Rectsv,
	__glXDisp_TexCoord1dv,
	__glXDisp_TexCoord1fv,
	__glXDisp_TexCoord1iv,
	__glXDisp_TexCoord1sv,
	__glXDisp_TexCoord2dv,
	__glXDisp_TexCoord2fv,
	__glXDisp_TexCoord2iv,
	__glXDisp_TexCoord2sv,
	__glXDisp_TexCoord3dv,
	__glXDisp_TexCoord3fv,
	__glXDisp_TexCoord3iv,
	__glXDisp_TexCoord3sv,
	__glXDisp_TexCoord4dv,
	__glXDisp_TexCoord4fv,
	__glXDisp_TexCoord4iv,
	__glXDisp_TexCoord4sv,
	__glXDisp_Vertex2dv,
	__glXDisp_Vertex2fv,
	__glXDisp_Vertex2iv,
	__glXDisp_Vertex2sv,
	__glXDisp_Vertex3dv,
	__glXDisp_Vertex3fv,
	__glXDisp_Vertex3iv,
	__glXDisp_Vertex3sv,
	__glXDisp_Vertex4dv,
	__glXDisp_Vertex4fv,
	__glXDisp_Vertex4iv,
	__glXDisp_Vertex4sv,
	__glXDisp_ClipPlane,
	__glXDisp_ColorMaterial,
	__glXDisp_CullFace,
	__glXDisp_Fogf,
	__glXDisp_Fogfv,
	__glXDisp_Fogi,
	__glXDisp_Fogiv,
	__glXDisp_FrontFace,
	__glXDisp_Hint,
	__glXDisp_Lightf,
	__glXDisp_Lightfv,
	__glXDisp_Lighti,
	__glXDisp_Lightiv,
	__glXDisp_LightModelf,
	__glXDisp_LightModelfv,
	__glXDisp_LightModeli,
	__glXDisp_LightModeliv,
	__glXDisp_LineStipple,
	__glXDisp_LineWidth,
	__glXDisp_Materialf,
	__glXDisp_Materialfv,
	__glXDisp_Materiali,
	__glXDisp_Materialiv,
	__glXDisp_PointSize,
	__glXDisp_PolygonMode,
	__glXDisp_PolygonStipple,
	__glXDisp_Scissor,
	__glXDisp_ShadeModel,
	__glXDisp_TexParameterf,
	__glXDisp_TexParameterfv,
	__glXDisp_TexParameteri,
	__glXDisp_TexParameteriv,
	__glXDisp_TexImage1D,
	__glXDisp_TexImage2D,
	__glXDisp_TexEnvf,
	__glXDisp_TexEnvfv,
	__glXDisp_TexEnvi,
	__glXDisp_TexEnviv,
	__glXDisp_TexGend,
	__glXDisp_TexGendv,
	__glXDisp_TexGenf,
	__glXDisp_TexGenfv,
	__glXDisp_TexGeni,
	__glXDisp_TexGeniv,
	__glXDisp_InitNames,
	__glXDisp_LoadName,
	__glXDisp_PassThrough,
	__glXDisp_PopName,
	__glXDisp_PushName,
	__glXDisp_DrawBuffer,
	__glXDisp_Clear,
	__glXDisp_ClearAccum,
	__glXDisp_ClearIndex,
	__glXDisp_ClearColor,
	__glXDisp_ClearStencil,
	__glXDisp_ClearDepth,
	__glXDisp_StencilMask,
	__glXDisp_ColorMask,
	__glXDisp_DepthMask,
	__glXDisp_IndexMask,
	__glXDisp_Accum,
	__glXDisp_Disable,
	__glXDisp_Enable,
	__glXNoSuchRenderOpcode,
	__glXDisp_PopAttrib,
	__glXDisp_PushAttrib,
	__glXDisp_Map1d,
	__glXDisp_Map1f,
	__glXDisp_Map2d,
	__glXDisp_Map2f,
	__glXDisp_MapGrid1d,
	__glXDisp_MapGrid1f,
	__glXDisp_MapGrid2d,
	__glXDisp_MapGrid2f,
	__glXDisp_EvalCoord1dv,
	__glXDisp_EvalCoord1fv,
	__glXDisp_EvalCoord2dv,
	__glXDisp_EvalCoord2fv,
	__glXDisp_EvalMesh1,
	__glXDisp_EvalPoint1,
	__glXDisp_EvalMesh2,
	__glXDisp_EvalPoint2,
	__glXDisp_AlphaFunc,
	__glXDisp_BlendFunc,
	__glXDisp_LogicOp,
	__glXDisp_StencilFunc,
	__glXDisp_StencilOp,
	__glXDisp_DepthFunc,
	__glXDisp_PixelZoom,
	__glXDisp_PixelTransferf,
	__glXDisp_PixelTransferi,
	__glXDisp_PixelMapfv,
	__glXDisp_PixelMapuiv,
	__glXDisp_PixelMapusv,
	__glXDisp_ReadBuffer,
	__glXDisp_CopyPixels,
	__glXDisp_DrawPixels,
	__glXDisp_DepthRange,
	__glXDisp_Frustum,
	__glXDisp_LoadIdentity,
	__glXDisp_LoadMatrixf,
	__glXDisp_LoadMatrixd,
	__glXDisp_MatrixMode,
	__glXDisp_MultMatrixf,
	__glXDisp_MultMatrixd,
	__glXDisp_Ortho,
	__glXDisp_PopMatrix,
	__glXDisp_PushMatrix,
	__glXDisp_Rotated,
	__glXDisp_Rotatef,
	__glXDisp_Scaled,
	__glXDisp_Scalef,
	__glXDisp_Translated,
	__glXDisp_Translatef,
	__glXDisp_Viewport,
	__glXDisp_PolygonOffset,
	__glXDisp_DrawArrays,
	__glXDisp_Indexubv,
	__glXDisp_ColorSubTable,
	__glXDisp_CopyColorSubTable,
	__glXDisp_ActiveTextureARB,
	__glXDisp_MultiTexCoord1dvARB,
	__glXDisp_MultiTexCoord1fvARB,
	__glXDisp_MultiTexCoord1ivARB,
	__glXDisp_MultiTexCoord1svARB,
	__glXDisp_MultiTexCoord2dvARB,
	__glXDisp_MultiTexCoord2fvARB,
	__glXDisp_MultiTexCoord2ivARB,
	__glXDisp_MultiTexCoord2svARB,
	__glXDisp_MultiTexCoord3dvARB,
	__glXDisp_MultiTexCoord3fvARB,
	__glXDisp_MultiTexCoord3ivARB,
	__glXDisp_MultiTexCoord3svARB,
	__glXDisp_MultiTexCoord4dvARB,
	__glXDisp_MultiTexCoord4fvARB,
	__glXDisp_MultiTexCoord4ivARB,
	__glXDisp_MultiTexCoord4svARB, /* 213 */
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode, /* 220 */
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode,
	__glXDisp_SampleCoverageARB,
#ifndef MISSING_GL_EXTS
	__glXDisp_WindowPos3fARB /* 230 */
#else
	__glXNoSuchRenderOpcode
#endif
};

__GLXdispatchSingleProcPtr __glXSwapSingleTable[__GLX_SINGLE_TABLE_SIZE] = {
	__glXNoSuchSingleOpcode,
	__glXSwapRender,
	__glXSwapRenderLarge,
	__glXSwapCreateContext,
	__glXSwapDestroyContext,
	__glXSwapMakeCurrent,
	__glXSwapIsDirect,
	__glXSwapQueryVersion,
	__glXSwapWaitGL,
	__glXSwapWaitX,
	__glXSwapCopyContext,
	__glXSwapSwapBuffers,
	__glXSwapUseXFont,
	__glXSwapCreateGLXPixmap,
	__glXSwapGetVisualConfigs,
	__glXSwapDestroyGLXPixmap,
	__glXSwapVendorPrivate,
	__glXSwapVendorPrivateWithReply,
	__glXSwapQueryExtensionsString,
	__glXSwapQueryServerString,
	__glXSwapClientInfo,
	__glXSwapGetFBConfigs,
	__glXSwapCreatePixmap,
	__glXSwapDestroyGLXPixmap, /* glXDestroyPixmap */
	__glXSwapCreateNewContext,
	__glXNoSuchSingleOpcode, /* glXQueryContext */
	__glXSwapMakeContextCurrent,
	__glXNoSuchSingleOpcode, /* glXCreatePbuffer */
	__glXNoSuchSingleOpcode, /* glXDestroyPbuffer */
	__glXNoSuchSingleOpcode, /* glXGetDrawableAttributes */
	__glXNoSuchSingleOpcode, /* glXChangeDrawableAttributes */
	__glXNoSuchSingleOpcode, /* glXCreateWindow */
	__glXNoSuchSingleOpcode, /* glXDestroyWindow */
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXNoSuchSingleOpcode,
	__glXDispSwap_NewList,
	__glXDispSwap_EndList,
	__glXDispSwap_DeleteLists,
	__glXDispSwap_GenLists,
	__glXDispSwap_FeedbackBuffer,
	__glXDispSwap_SelectBuffer,
	__glXDispSwap_RenderMode,
	__glXDispSwap_Finish,
	__glXDispSwap_PixelStoref,
	__glXDispSwap_PixelStorei,
	__glXDispSwap_ReadPixels,
	__glXDispSwap_GetBooleanv,
	__glXDispSwap_GetClipPlane,
	__glXDispSwap_GetDoublev,
	__glXDispSwap_GetError,
	__glXDispSwap_GetFloatv,
	__glXDispSwap_GetIntegerv,
	__glXDispSwap_GetLightfv,
	__glXDispSwap_GetLightiv,
	__glXDispSwap_GetMapdv,
	__glXDispSwap_GetMapfv,
	__glXDispSwap_GetMapiv,
	__glXDispSwap_GetMaterialfv,
	__glXDispSwap_GetMaterialiv,
	__glXDispSwap_GetPixelMapfv,
	__glXDispSwap_GetPixelMapuiv,
	__glXDispSwap_GetPixelMapusv,
	__glXDispSwap_GetPolygonStipple,
	__glXDispSwap_GetString,
	__glXDispSwap_GetTexEnvfv,
	__glXDispSwap_GetTexEnviv,
	__glXDispSwap_GetTexGendv,
	__glXDispSwap_GetTexGenfv,
	__glXDispSwap_GetTexGeniv,
	__glXDispSwap_GetTexImage,
	__glXDispSwap_GetTexParameterfv,
	__glXDispSwap_GetTexParameteriv,
	__glXDispSwap_GetTexLevelParameterfv,
	__glXDispSwap_GetTexLevelParameteriv,
	__glXDispSwap_IsEnabled,
	__glXDispSwap_IsList,
	__glXDispSwap_Flush,
	__glXDispSwap_AreTexturesResident,
	__glXDispSwap_DeleteTextures,
	__glXDispSwap_GenTextures,
	__glXDispSwap_IsTexture,
	__glXDispSwap_GetColorTable,
	__glXDispSwap_GetColorTableParameterfv,
	__glXDispSwap_GetColorTableParameteriv,
	__glXDispSwap_GetConvolutionFilter,
	__glXDispSwap_GetConvolutionParameterfv,
	__glXDispSwap_GetConvolutionParameteriv,
	__glXDispSwap_GetSeparableFilter,
	__glXDispSwap_GetHistogram,
	__glXDispSwap_GetHistogramParameterfv,
	__glXDispSwap_GetHistogramParameteriv,
	__glXDispSwap_GetMinmax,
	__glXDispSwap_GetMinmaxParameterfv,
	__glXDispSwap_GetMinmaxParameteriv,
};

__GLXdispatchRenderProcPtr __glXSwapRenderTable[__GLX_RENDER_TABLE_SIZE] = {
	__glXNoSuchRenderOpcode,
	__glXDispSwap_CallList,
	__glXDispSwap_CallLists,
	__glXDispSwap_ListBase,
	__glXDispSwap_Begin,
	__glXDispSwap_Bitmap,
	__glXDispSwap_Color3bv,
	__glXDispSwap_Color3dv,
	__glXDispSwap_Color3fv,
	__glXDispSwap_Color3iv,
	__glXDispSwap_Color3sv,
	__glXDispSwap_Color3ubv,
	__glXDispSwap_Color3uiv,
	__glXDispSwap_Color3usv,
	__glXDispSwap_Color4bv,
	__glXDispSwap_Color4dv,
	__glXDispSwap_Color4fv,
	__glXDispSwap_Color4iv,
	__glXDispSwap_Color4sv,
	__glXDispSwap_Color4ubv,
	__glXDispSwap_Color4uiv,
	__glXDispSwap_Color4usv,
	__glXDispSwap_EdgeFlagv,
	__glXDispSwap_End,
	__glXDispSwap_Indexdv,
	__glXDispSwap_Indexfv,
	__glXDispSwap_Indexiv,
	__glXDispSwap_Indexsv,
	__glXDispSwap_Normal3bv,
	__glXDispSwap_Normal3dv,
	__glXDispSwap_Normal3fv,
	__glXDispSwap_Normal3iv,
	__glXDispSwap_Normal3sv,
	__glXDispSwap_RasterPos2dv,
	__glXDispSwap_RasterPos2fv,
	__glXDispSwap_RasterPos2iv,
	__glXDispSwap_RasterPos2sv,
	__glXDispSwap_RasterPos3dv,
	__glXDispSwap_RasterPos3fv,
	__glXDispSwap_RasterPos3iv,
	__glXDispSwap_RasterPos3sv,
	__glXDispSwap_RasterPos4dv,
	__glXDispSwap_RasterPos4fv,
	__glXDispSwap_RasterPos4iv,
	__glXDispSwap_RasterPos4sv,
	__glXDispSwap_Rectdv,
	__glXDispSwap_Rectfv,
	__glXDispSwap_Rectiv,
	__glXDispSwap_Rectsv,
	__glXDispSwap_TexCoord1dv,
	__glXDispSwap_TexCoord1fv,
	__glXDispSwap_TexCoord1iv,
	__glXDispSwap_TexCoord1sv,
	__glXDispSwap_TexCoord2dv,
	__glXDispSwap_TexCoord2fv,
	__glXDispSwap_TexCoord2iv,
	__glXDispSwap_TexCoord2sv,
	__glXDispSwap_TexCoord3dv,
	__glXDispSwap_TexCoord3fv,
	__glXDispSwap_TexCoord3iv,
	__glXDispSwap_TexCoord3sv,
	__glXDispSwap_TexCoord4dv,
	__glXDispSwap_TexCoord4fv,
	__glXDispSwap_TexCoord4iv,
	__glXDispSwap_TexCoord4sv,
	__glXDispSwap_Vertex2dv,
	__glXDispSwap_Vertex2fv,
	__glXDispSwap_Vertex2iv,
	__glXDispSwap_Vertex2sv,
	__glXDispSwap_Vertex3dv,
	__glXDispSwap_Vertex3fv,
	__glXDispSwap_Vertex3iv,
	__glXDispSwap_Vertex3sv,
	__glXDispSwap_Vertex4dv,
	__glXDispSwap_Vertex4fv,
	__glXDispSwap_Vertex4iv,
	__glXDispSwap_Vertex4sv,
	__glXDispSwap_ClipPlane,
	__glXDispSwap_ColorMaterial,
	__glXDispSwap_CullFace,
	__glXDispSwap_Fogf,
	__glXDispSwap_Fogfv,
	__glXDispSwap_Fogi,
	__glXDispSwap_Fogiv,
	__glXDispSwap_FrontFace,
	__glXDispSwap_Hint,
	__glXDispSwap_Lightf,
	__glXDispSwap_Lightfv,
	__glXDispSwap_Lighti,
	__glXDispSwap_Lightiv,
	__glXDispSwap_LightModelf,
	__glXDispSwap_LightModelfv,
	__glXDispSwap_LightModeli,
	__glXDispSwap_LightModeliv,
	__glXDispSwap_LineStipple,
	__glXDispSwap_LineWidth,
	__glXDispSwap_Materialf,
	__glXDispSwap_Materialfv,
	__glXDispSwap_Materiali,
	__glXDispSwap_Materialiv,
	__glXDispSwap_PointSize,
	__glXDispSwap_PolygonMode,
	__glXDispSwap_PolygonStipple,
	__glXDispSwap_Scissor,
	__glXDispSwap_ShadeModel,
	__glXDispSwap_TexParameterf,
	__glXDispSwap_TexParameterfv,
	__glXDispSwap_TexParameteri,
	__glXDispSwap_TexParameteriv,
	__glXDispSwap_TexImage1D,
	__glXDispSwap_TexImage2D,
	__glXDispSwap_TexEnvf,
	__glXDispSwap_TexEnvfv,
	__glXDispSwap_TexEnvi,
	__glXDispSwap_TexEnviv,
	__glXDispSwap_TexGend,
	__glXDispSwap_TexGendv,
	__glXDispSwap_TexGenf,
	__glXDispSwap_TexGenfv,
	__glXDispSwap_TexGeni,
	__glXDispSwap_TexGeniv,
	__glXDispSwap_InitNames,
	__glXDispSwap_LoadName,
	__glXDispSwap_PassThrough,
	__glXDispSwap_PopName,
	__glXDispSwap_PushName,
	__glXDispSwap_DrawBuffer,
	__glXDispSwap_Clear,
	__glXDispSwap_ClearAccum,
	__glXDispSwap_ClearIndex,
	__glXDispSwap_ClearColor,
	__glXDispSwap_ClearStencil,
	__glXDispSwap_ClearDepth,
	__glXDispSwap_StencilMask,
	__glXDispSwap_ColorMask,
	__glXDispSwap_DepthMask,
	__glXDispSwap_IndexMask,
	__glXDispSwap_Accum,
	__glXDispSwap_Disable,
	__glXDispSwap_Enable,
	__glXNoSuchRenderOpcode,
	__glXDispSwap_PopAttrib,
	__glXDispSwap_PushAttrib,
	__glXDispSwap_Map1d,
	__glXDispSwap_Map1f,
	__glXDispSwap_Map2d,
	__glXDispSwap_Map2f,
	__glXDispSwap_MapGrid1d,
	__glXDispSwap_MapGrid1f,
	__glXDispSwap_MapGrid2d,
	__glXDispSwap_MapGrid2f,
	__glXDispSwap_EvalCoord1dv,
	__glXDispSwap_EvalCoord1fv,
	__glXDispSwap_EvalCoord2dv,
	__glXDispSwap_EvalCoord2fv,
	__glXDispSwap_EvalMesh1,
	__glXDispSwap_EvalPoint1,
	__glXDispSwap_EvalMesh2,
	__glXDispSwap_EvalPoint2,
	__glXDispSwap_AlphaFunc,
	__glXDispSwap_BlendFunc,
	__glXDispSwap_LogicOp,
	__glXDispSwap_StencilFunc,
	__glXDispSwap_StencilOp,
	__glXDispSwap_DepthFunc,
	__glXDispSwap_PixelZoom,
	__glXDispSwap_PixelTransferf,
	__glXDispSwap_PixelTransferi,
	__glXDispSwap_PixelMapfv,
	__glXDispSwap_PixelMapuiv,
	__glXDispSwap_PixelMapusv,
	__glXDispSwap_ReadBuffer,
	__glXDispSwap_CopyPixels,
	__glXDispSwap_DrawPixels,
	__glXDispSwap_DepthRange,
	__glXDispSwap_Frustum,
	__glXDispSwap_LoadIdentity,
	__glXDispSwap_LoadMatrixf,
	__glXDispSwap_LoadMatrixd,
	__glXDispSwap_MatrixMode,
	__glXDispSwap_MultMatrixf,
	__glXDispSwap_MultMatrixd,
	__glXDispSwap_Ortho,
	__glXDispSwap_PopMatrix,
	__glXDispSwap_PushMatrix,
	__glXDispSwap_Rotated,
	__glXDispSwap_Rotatef,
	__glXDispSwap_Scaled,
	__glXDispSwap_Scalef,
	__glXDispSwap_Translated,
	__glXDispSwap_Translatef,
	__glXDispSwap_Viewport,
	__glXDispSwap_PolygonOffset,
	__glXDispSwap_DrawArrays,
	__glXDispSwap_Indexubv,
	__glXDispSwap_ColorSubTable,
	__glXDispSwap_CopyColorSubTable,
	__glXDispSwap_ActiveTextureARB,
	__glXDispSwap_MultiTexCoord1dvARB,
	__glXDispSwap_MultiTexCoord1fvARB,
	__glXDispSwap_MultiTexCoord1ivARB,
	__glXDispSwap_MultiTexCoord1svARB,
	__glXDispSwap_MultiTexCoord2dvARB,
	__glXDispSwap_MultiTexCoord2fvARB,
	__glXDispSwap_MultiTexCoord2ivARB,
	__glXDispSwap_MultiTexCoord2svARB,
	__glXDispSwap_MultiTexCoord3dvARB,
	__glXDispSwap_MultiTexCoord3fvARB,
	__glXDispSwap_MultiTexCoord3ivARB,
	__glXDispSwap_MultiTexCoord3svARB,
	__glXDispSwap_MultiTexCoord4dvARB,
	__glXDispSwap_MultiTexCoord4fvARB,
	__glXDispSwap_MultiTexCoord4ivARB,
	__glXDispSwap_MultiTexCoord4svARB, /* 213 */
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode, /* 220 */
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode,
	__glXNoSuchRenderOpcode,
	__glXDispSwap_SampleCoverageARB,
#ifndef MISSING_GL_EXTS
	__glXDispSwap_WindowPos3fARB /* 230 */
#else
	__glXNoSuchRenderOpcode
#endif
};
