/*
 * Copyright (c) 1987 by the Regents of the University of California
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies.  The University of
 * California makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 */

/*
 * Moved here from mi to allow wrapping of lower level backing store functions.
 * -- 1997.10.27  Marc Aurele La France (tsi@xfree86.org)
 */

#ifndef _BSTORE_H_
#define _BSTORE_H_

#include "bstorestr.h"

#endif /* _BSTORE_H_ */
