/*
 * Mesa 3-D graphics library
 * Version:  6.3
 *
 * Copyright (C) 2005  Brian Paul   All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#if !defined SLANG_EXECUTE_H
#define SLANG_EXECUTE_H

#if defined __cplusplus
extern "C" {
#endif

#define SLANG_MACHINE_STACK_SIZE 1024

typedef struct slang_machine_
{
	GLuint ip;					/* instruction pointer, for flow control */
	GLuint sp;					/* stack pointer, for stack access */
	GLuint bp;					/* base pointer, for local variable access */
	GLuint kill;				/* discard the fragment */
	GLuint exit;				/* terminate the shader */
	union stack_
	{
		GLfloat _float[SLANG_MACHINE_STACK_SIZE];
		GLfloat *_floatp[SLANG_MACHINE_STACK_SIZE];
		GLuint _addr[SLANG_MACHINE_STACK_SIZE];
		GLuint *_addrp[SLANG_MACHINE_STACK_SIZE];
	} stack;
} slang_machine;

int _slang_execute (const slang_assembly_file *);

#ifdef __cplusplus
}
#endif

#endif

